@@ -303,6 +303,7 @@ static struct less_charset_entry less_charset_table[] = {
 	{ "UTF-8",		"utf-8",	NULL		},
 
 #ifdef MULTIBYTE_GROFF
+	{ "CP1251",		"windows",	NULL		},
 	{ "EUC-JP",		"iso8859",	"japanese-ujis"	},
 	{ "KOI8-R",		"koi8-r",	NULL		},
 	/* close enough? */
@@ -314,6 +315,58 @@ static struct less_charset_entry less_charset_table[] = {
 
 static const char *fallback_less_charset = "iso8859";
 
+/* Encoding conversions from groff-1.20/src/preproc/preconv/preconv.cpp.
+ * I've only included those not already recognised by GNU libiconv.
+ */
+struct conversion_entry {
+	const char *from;
+	const char *to;
+};
+
+static struct conversion_entry conversion_table[] = {
+	{ "chinese-big5",			"Big5" },
+	{ "chinese-euc",			"GB2312" },
+	{ "chinese-iso-8bit",			"GB2312" },
+	{ "cn-gb-2312",				"GB2312" },
+	{ "cp878",				"KOI8-R" },
+	{ "cyrillic-iso-8bit",			"ISO-8859-5" },
+	{ "cyrillic-koi8",			"KOI8-R" },
+	{ "euc-china",				"GB2312" },
+	{ "euc-japan",				"EUC-JP" },
+	{ "euc-japan-1990",			"EUC-JP" },
+	{ "euc-kr",				"EUC-KR" },
+	{ "greek-iso-8bit",			"ISO-8859-7" },
+	{ "iso-latin-1",			"ISO-8859-1" },
+	{ "iso-latin-2",			"ISO-8859-2" },
+	{ "iso-latin-5",			"ISO-8859-9" },
+	{ "iso-latin-7",			"ISO-8859-13" },
+	{ "iso-latin-9",			"ISO-8859-15" },
+	{ "japanese-iso-8bit",			"EUC-JP" },
+	{ "japanese-euc",			"EUC-JP" },
+	{ "jis8",				"EUC-JP" },
+	{ "korean-euc",				"EUC-KR" },
+	{ "korean-iso-8bit",			"EUC-KR" },
+	{ "latin-0",				"ISO-8859-15" },
+	{ "latin-1",				"ISO-8859-1" },
+	{ "latin-2",				"ISO-8859-2" },
+	{ "latin-5",				"ISO-8859-9" },
+	{ "latin-7",				"ISO-8859-13" },
+	{ "mule-utf-16",			"UTF-16" },
+	{ "mule-utf-16be",			"UTF-16BE" },
+	{ "mule-utf-16-be",			"UTF-16BE" },
+	{ "mule-utf-16be-with-signature",	"UTF-16" },
+	{ "mule-utf-16le",			"UTF-16LE" },
+	{ "mule-utf-16-le",			"UTF-16LE" },
+	{ "mule-utf-16le-with-signature",	"UTF-16" },
+	{ "mule-utf-8",				"UTF-8" },
+	{ "utf-16-be",				"UTF-16BE" },
+	{ "utf-16be-with-signature",		"UTF-16" },
+	{ "utf-16-be-with-signature",		"UTF-16" },
+	{ "utf-16-le",				"UTF-16LE" },
+	{ "utf-16le-with-signature",		"UTF-16" },
+	{ "utf-16-le-with-signature",		"UTF-16" },
+	{ NULL,					NULL }
+};
 
 const char *groff_preconv = NULL;
 
@@ -702,9 +755,12 @@ const char *get_less_charset (const char *locale_charset)
 {
 	const struct less_charset_entry *entry;
 
-	for (entry = less_charset_table; entry->locale_charset; ++entry)
-		if (STREQ (entry->locale_charset, locale_charset))
-			return entry->less_charset;
+	if (locale_charset) {
+		for (entry = less_charset_table; entry->locale_charset;
+		     ++entry)
+			if (STREQ (entry->locale_charset, locale_charset))
+				return entry->less_charset;
+	}
 
 	return fallback_less_charset;
 }
@@ -716,13 +772,43 @@ const char *get_jless_charset (const char *locale_charset)
 {
 	const struct less_charset_entry *entry;
 
-	for (entry = less_charset_table; entry->locale_charset; ++entry)
-		if (STREQ (entry->locale_charset, locale_charset))
-			return entry->jless_charset;
+	if (locale_charset) {
+		for (entry = less_charset_table; entry->locale_charset;
+		     ++entry)
+			if (STREQ (entry->locale_charset, locale_charset))
+				return entry->jless_charset;
+	}
 
 	return NULL;
 }
 
+/* Convert Emacs-style coding tags to ones that libiconv understands. */
+static char *convert_encoding (char *encoding)
+{
+	size_t encoding_len = strlen (encoding);
+	const struct conversion_entry *entry;
+
+#define STRIP(s, l) do { \
+	if (encoding_len > (l) && \
+	    !strcasecmp (encoding + encoding_len - (l), (s))) \
+		encoding[encoding_len - (l)] = '\0'; \
+} while (0)
+
+	STRIP ("-dos", 4);
+	STRIP ("-mac", 4);
+	STRIP ("-unix", 5);
+
+#undef STRIP
+
+	for (entry = conversion_table; entry->from; ++entry)
+		if (!strcasecmp (entry->from, encoding)) {
+			free (encoding);
+			return xstrdup (entry->to);
+		}
+
+	return encoding;
+}
+
 /* Inspect the first line of data in a pipeline for preprocessor encoding
  * declarations.
  */
@@ -749,21 +835,31 @@ char *check_preprocessor_encoding (pipeline *p)
 
 	if (directive && strstr (directive, "-*-")) {
 		const char *pp_search = strstr (directive, "-*-") + 3;
-		while (*pp_search == ' ')
-			++pp_search;
-		if (STRNEQ (pp_search, "coding:", 7)) {
-			const char *pp_encoding_allow;
-			size_t pp_encoding_len;
-			pp_search += 7;
+		while (pp_search && *pp_search) {
 			while (*pp_search == ' ')
 				++pp_search;
-			pp_encoding_allow = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
-					    "abcdefghijklmnopqrstuvwxyz"
-					    "0123456789-_/:.()";
-			pp_encoding_len = strspn (pp_search,
-						  pp_encoding_allow);
-			pp_encoding = xstrndup (pp_search, pp_encoding_len);
-			debug ("preprocessor encoding: %s\n", pp_encoding);
+			if (STRNEQ (pp_search, "coding:", 7)) {
+				const char *pp_encoding_allow;
+				size_t pp_encoding_len;
+				pp_search += 7;
+				while (*pp_search == ' ')
+					++pp_search;
+				pp_encoding_allow = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
+						    "abcdefghijklmnopqrstuvwxyz"
+						    "0123456789-_/:.()";
+				pp_encoding_len = strspn (pp_search,
+							  pp_encoding_allow);
+				pp_encoding = xstrndup (pp_search,
+							pp_encoding_len);
+				pp_encoding = convert_encoding (pp_encoding);
+				debug ("preprocessor encoding: %s\n",
+				       pp_encoding);
+				break;
+			} else {
+				pp_search = strchr (pp_search, ';');
+				if (pp_search)
+					++pp_search;
+			}
 		}
 	}
 #endif /* PP_COOKIE */
